import torch
from torch.distributions import constraints

from .survival_distribution import SurvivalDistribution


class Exponential(SurvivalDistribution, torch.distributions.Exponential):
    """Exponential distribution with rate parametrization.

    References:
        https://en.wikipedia.org/wiki/Exponential_distribution


    Args:
        rate: Rate parameter, equal to 1 / scale.
    """

    support = constraints.positive

    def logsf(self, value):
        if self._validate_args:
            self._validate_sample(value)
        return -self.rate * value

    def log_hazard(self, value):
        if self._validate_args:
            self._validate_sample(value)
        return self.rate.log().expand_as(value)

    def isf(self, u):
        if self._validate_args:
            assert constraints.unit_interval.check(u).all()
        return -torch.log(u) / self.rate

    def icdf(self, u):
        return self.isf(1.0 - u)

    def _new_tensor(self, shape):
        return self.rate.new(shape)



import torch
import torch.nn.functional as F
from torch.distributions import constraints
from torch.distributions.utils import broadcast_all

from .survival_distribution import SurvivalDistribution


class Logistic(SurvivalDistribution):
    arg_constraints = {"loc": constraints.real, "scale": constraints.positive}
    support = constraints.real

    def __init__(self, loc, scale, validate_args=None):
        self.loc, self.scale = broadcast_all(loc, scale)
        batch_shape = self.loc.shape
        super().__init__(batch_shape, validate_args=validate_args)

    def sf(self, value):
        if self._validate_args:
            self._validate_sample(value)
        z = (value - self.loc) / self.scale
        return torch.sigmoid(-z)

    def logsf(self, value):
        if self._validate_args:
            self._validate_sample(value)
        z = (value - self.loc) / self.scale
        return F.logsigmoid(-z)

    def cdf(self, value):
        if self._validate_args:
            self._validate_sample(value)
        z = (value - self.loc) / self.scale
        return torch.sigmoid(z)

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)
        z = (value - self.loc) / self.scale
        return z - self.scale.log() - 2 * F.softplus(z)

    def isf(self, u):
        if self._validate_args:
            assert constraints.unit_interval.check(u).all()
        z = (-u).log1p() - u.log()
        return self.scale * z + self.loc

    def _new_tensor(self, shape):
        return self.loc.new(shape)



from .exponential import Exponential
from .log_logistic import LogLogistic
from .log_normal import LogNormal
from .logistic import Logistic
from .mixture_same_family import MixtureSameFamily
from .normal import Normal
from .survival_distribution import SurvivalDistribution
from .transformed_distribution import TransformedDistribution
from .uniform import Uniform
from .weibull import Weibull



import torch

from .survival_distribution import SurvivalDistribution


class TransformedDistribution(
    torch.distributions.TransformedDistribution, SurvivalDistribution
):
    base_dist: SurvivalDistribution

    def __init__(self, base_distribution, transforms, validate_args=None):
        super().__init__(base_distribution, transforms, validate_args=validate_args)
        sign = 1
        for transform in self.transforms:
            sign = sign * transform.sign
        if sign not in (-1, 1):
            raise ValueError("Each transform must have sign +1 or -1")
        # sign equals to -1 if the transformation is decreasing, +1 if increasing
        self.sign = int(sign)

    def logsf(self, value):
        x = value
        for transform in self.transforms[::-1]:
            x = transform.inv(x)
        if self._validate_args:
            self.base_dist._validate_sample(x)
        if self.sign == 1:
            return self.base_dist.logsf(x)
        else:
            return self.base_dist.logcdf(x)

    def logcdf(self, value):
        x = value
        for transform in self.transforms[::-1]:
            x = transform.inv(x)
        if self._validate_args:
            self.base_dist._validate_sample(x)
        if self.sign == 1:
            return self.base_dist.logcdf(x)
        else:
            return self.base_dist.logsf(x)

    def log_hazard(self, value):
        # The naive implementation `log_prob(value) - logsf(value)` goes through all the
        # transformations twice. This implementation only goes through the transforms
        # once
        log_hazard = 0.0
        y = value
        for transform in reversed(self.transforms):
            x = transform.inv(y)
            log_hazard = log_hazard - transform.log_abs_det_jacobian(x, y)
            y = x

        log_hazard = log_hazard + self.base_dist.log_hazard(y)
        return log_hazard

    def isf(self, u):
        x = self.base_dist.isf(u)
        for transform in self.transforms:
            x = transform(x)
        return x

    def _new_tensor(self, shape):
        return self.base_dist._new_tensor(shape)



from setuptools import find_packages, setup

setup(
    name="survival_distributions",
    author="Oleksandr Shchur",
    description="Extended functionality for univariate probability distributions in PyTorch",
    version="0.0.3",
    license="MIT",
    author_email="oleks.shchur@gmail.com",
    url="https://github.com/shchur/survival_distributions",
    packages=find_packages(),
    python_requires=">=3.7",
    install_requires=["torch>=1.10"],
)
